from .IConfigurable import IConfigurable
from .IProcessor import IProcessor
from .IResourceAware import IResourceAware
from .LoggableMixin import LoggableMixin

from .IDataService import IDataService



def duckyinterface(target_cls):
    @classmethod
    def check(cls, subcls):
        if cls is not target_cls:
            return super(cls).__subclasshook__(subcls)

        if not hasattr(target_cls, '__abstractmethods__'): return True

        methods = target_cls.__abstractmethods__
        res = True
        for m in methods:
            res &= hasattr(subcls, m) and callable(getattr(subcls, m))

        return res

    setattr(target_cls, '__subclasshook__', check)

    return target_cls




from typing import List, Any, Tuple

from .IDataProcess import IDataProcess


class TranslateDataProcess(IDataProcess):
    """ DataProcess to change data fields names and preform tranformations on the
        values. Config dict foramat:
        {
            'exclude': list, # List of input fields to exclude
            'translations': dict, # Dict where keys are field names on
                input (generated data of DataInput) and values are a new name
                for that field
            'transformations': dict # Dict where keys are field names and
                values can be a function to execute on the value or a constat
                value to be asigned to that field
        }
    """


    def __init__(self, config: dict=None):
        super().__init__()
        self._config = config

        self._exclude = []
        self._translations = {}
        self._transformations = {}

        if config: self.configure(config)

    def configure(self, config: dict):
        self._exclude = config.get('exclude', self._exclude)
        self._translations = config.get('translations', self._translations)
        self._transformations = config.get('transformations',
                                           self._transformations)

    def process_one(self, data: dict) -> dict:
        """ Process one data entry.

        Args:
            data (dict): Data to process.

        Returns:
            dict: Processed data.
        """
        p_data = {}
        for k, v in data.items():
            if k in self._exclude: continue
            p_key, p_val = self._process_field(k, v)
            p_data[p_key] = p_val

        return p_data

    def process_all(self, data: List[dict]) -> List[dict]:
        """ Process all data entries.

        Args:
            data (List[dict]): List of data to be processed.

        Returns:
            List[dict]: List with data processed.
        """
        return [self.process_one(d) for d in data]


    def _process_field(self, key, value) -> Tuple[Any, Any]:
        r_key = self._translations.get(key, key)

        if not key in self._transformations:
            return r_key, value

        r_value = self._transformations.get(key)
        if callable(r_value):
            r_value = r_value(value)

        return r_key, r_value



from typing import List
from abc import ABCMeta, abstractmethod

from transpydata.config import IDataService
from transpydata.util.decorators import duckyinterface


@duckyinterface
class IDataProcess(IDataService, metaclass=ABCMeta):


    def __init__(self):
        super().__init__()

    def process_one_method_name(self) -> str:
        return 'process_one'

    def process_all_method_name(self) -> str:
        return 'process_all'

    @abstractmethod
    def process_one(self, data: dict) -> dict:
        """ Process one data entry.

        Args:
            data (dict): Data to process.

        Returns:
            dict: Processed data.
        """
        raise NotImplementedError

    @abstractmethod
    def process_all(self, data: List[dict]) -> List[dict]:
        """ Process all data entries.

        Args:
            data (List[dict]): List of data to be processed.

        Returns:
            List[dict]: List with data processed.
        """
        raise NotImplementedError

    def initialize(self):
        super().initialize()

    def dispose(self):
        pass



from logging import Logger


class LoggableMixin():
    def __init__(self):
        super().__init__()
        self._logger = None # type: Logger

    @property
    def logger(self):
        return self._logger

    @logger.setter
    def logger(self, logger: Logger):
        self._logger = logger




from typing import List
from abc import ABCMeta, abstractmethod

from transpydata.config import IDataService
from transpydata.util.decorators import duckyinterface


@duckyinterface
class IDataInput(IDataService, metaclass=ABCMeta):

    def __init__(self):
        super().__init__()

    def process_one_method_name(self) -> str:
        return 'get_one'

    def process_all_method_name(self) -> str:
        return 'get_all'

    @abstractmethod
    def get_one(self, data: dict) -> dict:
        """ Get one data input entry. User can pass data to parametrize the item
            he wants each time.

        Args:
            data (dict): Data to parametrize/query the data input.

        Returns:
            dict: Data entry
        """
        raise NotImplementedError

    @abstractmethod
    def get_all(self) -> List[dict]:
        """ Get all input data at once. The arguments needed to perform this
            action should be passed through configuration.

        Returns:
            List[dict]: List of data entries.
        """
        raise NotImplementedError

    def initialize(self):
        super().initialize()

    def dispose(self):
        pass




from typing import Tuple, List
import json

import boto3
from botocore.exceptions import ClientError

from . import IDataOutput


class SQSDataOutput(IDataOutput):
    """ DataOutput that sends messages to AWS SQS.

    Config dict format:
    {
        'url': str, # SQS url
        'client_id': str, # AWS client id (optional, could be providad as env
            variable or default config file
            https://boto3.amazonaws.com/v1/documentation/api/latest/guide/configuration.html#using-environment-variables)
        'secret': str, # AWS client secret (optional, could be providad as env
            variable or default config file
            https://boto3.amazonaws.com/v1/documentation/api/latest/guide/configuration.html#using-environment-variables)
        'region': str, # AWS region (optional, could be providad as env
            variable or default config file
            https://boto3.amazonaws.com/v1/documentation/api/latest/guide/configuration.html#using-environment-variables)
        'session_token': str, # AWS session token (optional, could be providad as env
            variable or default config file
            https://boto3.amazonaws.com/v1/documentation/api/latest/guide/configuration.html#using-environment-variables)
        'endpoint_url': str, # AWS endpoint url config param. Mostly for local and development setups
        'attributes': dict, # Fields in data input that shoul go as attributes
            (key) and the name of the property (value)
    }

    """

    MAX_BATCH_SIZE = 10 # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sqs.html#SQS.Client.send_message_batch

    def __init__(self, config: dict = None):
        self.url = None
        self.attributes = {}

        self._client_id = ''
        self._secret = ''
        self._region = ''
        self._endpoint_url = ''
        self._session_token = ''
        self._sqs_client = None

        if config:
            self.configure(config)

    def configure(self, config: dict):
        self.url = config.get('url')
        if not self.url:
            raise RuntimeError("'url' parameter expected in configuration")

        self.attributes = config.get('attributes', self.attributes)
        self._client_id = config.get('client_id', self._client_id)
        self._secret = config.get('secret', self._secret)
        self._region = config.get('region', self._region)
        self._session_token = config.get('session_token', self._session_token)
        self._endpoint_url = config.get('endpoint_url', self._endpoint_url)

    def send_one(self, data: dict) -> dict:
        """ Send one data entry to SQS.

        Args:
            data (dict): SQS message data

        Returns:
            dict: Dict with process result. Result format:
            {
                'success': bool, # Whether messager has been posted correctly on SQS or not
                'message_id': str, # Message id in SQS (only if sucessful sending)
                'error': str, # Error in case of not successful sending
            }
        """
        sqs_client = self._get_sqs_client()
        sqs_msg = self._get_sqs_message_data(data)

        res = {}
        try:
            sqs_res = sqs_client.send_message(**sqs_msg)
            res = {'success': True, 'message_id': sqs_res['MessageId']}
        except ClientError as e:
            res = {'success': False, 'error': e}

        return res

    def send_all(self, data: List[dict]) -> List[dict]:
        """ Send a list of data entries to SQS.

        Args:
            data (List[dict]): SQS messages data

        Returns:
            List[dict]: List of dict entries with process result. Result format:
            {
                'success': bool, # Whether messager has been posted correctly on SQS or not
                'id': int, # Index of the message in the input data list
                'message_id': str, # Message id in SQS (only if sucessful sending)
                'error': str, # Error in case of not successful sending
            }
        """
        sqs_client = self._get_sqs_client()
        results = []
        for n in range(0, len(data), self.MAX_BATCH_SIZE):
            data_batch = data[n:n + self.MAX_BATCH_SIZE]
            sqs_data = self._get_sqs_messages_data(data_batch, n)

            sqs_res = sqs_client.send_message_batch(**sqs_data)
            results.extend(self._process_sqs_batch_result(sqs_res))

        return results

    def _process_sqs_batch_result(self, result: dict) -> List[dict]:
        proc_result = []
        if 'Successful' in result:
            for success_msg in result['Successful']:
                proc_result.append({
                    'success': True,
                    'id': success_msg['Id'],
                    'message_id': success_msg['MessageId']
                })

        if 'Failed' in result:
            for error_msg in result['Failed']:
                proc_result.append({
                    'success': False,
                    'id': error_msg['Id'],
                    'error': '[{}] {}'.format(error_msg['Code'], error_msg['Message'])
                })

        return proc_result

    def _get_sqs_messages_data(self, data: List[dict], start_id: int = 0) -> List[dict]:
        sqs_data = {'QueueUrl': self.url, 'Entries': []}
        for i, data_entry in enumerate(data, start=start_id):
            msg_data, attributes = self._process_data_and_attributes(data_entry)
            sqs_data['Entries'].append({
                'Id': str(i),
                'MessageBody': json.dumps(msg_data),
                'MessageAttributes': attributes
            })

        return sqs_data

    def _get_sqs_message_data(self, data: dict) -> dict:
        sqs_data = {'QueueUrl': self.url}

        msg_data, attributes = self._process_data_and_attributes(data)
        sqs_data['MessageBody'] = json.dumps(msg_data)
        sqs_data['MessageAttributes'] = attributes

        return sqs_data

    def _process_data_and_attributes(self, data: dict) -> Tuple[dict, dict]:
        """ Process data and split between message and attributes fields

        Args:
            data (dict): data

        Returns:
            Tuple[dict, dict]: First dict is messag data, second attributes data
        """
        if not self.attributes:
            return data, {}

        msg_data = {}
        msg_attributes = {}
        for key, val in data.items():
            if key in self.attributes.keys():
                msg_attributes[self.attributes[key]] = {
                    'StringValue': str(val),
                    'DataType': 'Number' if type(val) in [int, float] else 'String'
                }
            else:
                msg_data[key] = val

        return msg_data, msg_attributes

    def _get_sqs_client(self):
        if self._sqs_client:
            return self._sqs_client

        config = {}
        if self._client_id and self._secret:
            config['aws_access_key_id'] = self._client_id
            config['aws_secret_access_key'] = self._secret

        if self._region:
            config['region_name'] = self._region

        if self._session_token:
            config['aws_session_token'] = self._session_token

        if self._endpoint_url:
            config['endpoint_url'] = self._endpoint_url

        self._sqs_client = boto3.client('sqs', **config)

        return self._sqs_client



from abc import ABCMeta
from logging import getLogger, Logger, NullHandler

from transpydata.config import (
    IConfigurable, IProcessor, IResourceAware, LoggableMixin
)
from transpydata.util.decorators import duckyinterface


class IDataService(IProcessor, IConfigurable, IResourceAware,
                   LoggableMixin, metaclass=ABCMeta):

    def __init__(self):
        super().__init__()

    def initialize(self):
        if self.logger is None or not isinstance(self.logger, Logger):
            self.logger = getLogger('dummy')
            self.logger.addHandler(NullHandler())



import mysql.connector

from .IDataInput import IDataInput


class MysqlDataInput(IDataInput):
    """ DataInput to get data from Mysql. Config dict format:
    {
        'db_config': {
            'user': str,
            'password': str,
            'host': str,
            'port': int,
            'database': str
        },
        'get_one_query': str, # One item query. Interpolate variables with
            printf synthax. E.j: SELECT * FROM X WHERE id = %(id)s

        'get_all_query': str, # Query to get all items. Interpolate variables
            with printf synthax

        'all_query_params': dict, # Params to interpolate in query
    }

    """
    def __init__(self, config: dict = None):
        self._config = config

        self._db_config = None # type: dict
        self._db_connection = None # type: mysql.connection.MySQLConnection
        self._get_one_query = ''
        self._get_all_query = ''
        self._all_query_params = {}

        self._page_size = 0 # TODO: Not used for now

        if config: self.configure(config)

    def configure(self, config: dict):
        """Configure datainput. Refer to class doc to view the format

        Args:
            config (dict): configuration
        """
        self._config = config
        self._db_config = config['db_config']

        self._get_one_query = config.get('get_one_query', '')
        self._get_all_query = config.get('get_all_query', '')
        self._all_query_params = config.get('all_query_params', {})

        if not self._get_one_query and not self._get_all_query:
            raise RuntimeError(
                "'get_one_query' or 'get_all_query' needs to be provided"
            )

        self._page_size = config.get('page_size', None)

    def initialize(self):
        """ Create DB connection.
        """
        self._db_connection = mysql.connector.connect(**self._db_config)

    def dispose (self):
        """ Close DB connection.
        """
        if self._db_connection and self._db_connection.is_connected():
            self._db_connection.close()

    def _get_db_cursor(self) -> mysql.connector.cursor.MySQLCursor:
        return self._db_connection.cursor(dictionary=True)

    def get_all(self):
        return self._fetch_all_query(self._get_all_query, self._all_query_params)

    def get_one(self, data: dict):
        all_params = {**data, **self._all_query_params}
        return self._fetch_one_query(self._get_one_query, all_params)

    def _get_paged_query(self, query: str, num_page: int) -> str:

        # LIMIT
        offset = (num_page - 1) * self.page_size
        limit = f' LIMIT {offset}, {self.page_size}'

        return query + limit

    def _fetch_all_query(self, query: str, params: dict) -> dict:
        try:
            cursor = self._get_db_cursor()
            cursor.execute(query, params)

            return cursor.fetchall()

        finally:
            cursor.close()

        raise RuntimeError('Error performing query:\n'+query)

    def _fetch_one_query(self, query: str, params: dict) -> dict:
        try:
            cursor = self._get_db_cursor()
            cursor.execute(query, params)

            return cursor.fetchone()

        finally:
            cursor.close()

        raise RuntimeError('Error performing query:\n'+query)



__title__ = 'transpydata'
__description__ = 'Framework for managing migrations'
__url__ = 'https://github.com/salpreh/transpydata'
__version__ = '0.4.1'
__test_version__ = '0.2.0'
__author__ = 'salpreh'
__author_email__ = 'salva.perez46@gmail.com'
__license__ = 'Apache 2.0'



from typing import List

from .IDataProcess import IDataProcess

class NoneDataProcess(IDataProcess):
    """ A no operation data process. Will move the input from the DataIntput to
        DataOutput without changes. No need to configure.
    """
    def configure(self):
        pass
    def process_one(self, data: dict) -> dict:
        return data

    def process_all(self, data: List[dict]) -> List[dict]:
        return data



from typing import List
from abc import ABCMeta, abstractmethod
from logging import getLogger, Logger, NullHandler

from transpydata.config import IDataService
from transpydata.util.decorators import duckyinterface


@duckyinterface
class IDataOutput(IDataService, metaclass=ABCMeta):

    def __init__(self):
        super().__init__()

    def process_one_method_name(self) -> str:
        return 'send_one'

    def process_all_method_name(self) -> str:
        return 'send_all'

    @abstractmethod
    def send_one(self, data: dict) -> dict:
        """ Send one data entry to the implemented output destination.

        Args:
            data (dict): Data to send or configure the output
                (depends on implementation class)

        Returns:
            dict: Dict with details about send results.
        """
        raise NotImplementedError

    @abstractmethod
    def send_all(self, data: List[dict]) -> List[dict]:
        """ Send all data entries to the implemented output destination.

        Args:
            data (List[dict]): List of data to send or configure the output
                (depends on implementation class)

        Returns:
            List[dict]: List with details about send results.
        """
        raise NotImplementedError

    def initialize(self):
        super().initialize()

    def dispose(self):
        pass



from typing import List, Dict, Any
from uuid import uuid4
import json

import boto3
from botocore.exceptions import ClientError

from . import IDataInput


class SQSDataInput(IDataInput):
    """ DataOutput that sends messages to AWS SQS.

    Config dict format:
    {
        'url': str, # SQS url
        'client_id': str, # AWS client id (optional, could be providad as env
            variable or default config file
            https://boto3.amazonaws.com/v1/documentation/api/latest/guide/configuration.html#using-environment-variables)
        'secret': str, # AWS client secret (optional, could be providad as env
            variable or default config file
            https://boto3.amazonaws.com/v1/documentation/api/latest/guide/configuration.html#using-environment-variables)
        'region': str, # AWS region (optional, could be providad as env
            variable or default config file
            https://boto3.amazonaws.com/v1/documentation/api/latest/guide/configuration.html#using-environment-variables)
        'session_token': str, # AWS session token (optional, could be providad as env
            variable or default config file
            https://boto3.amazonaws.com/v1/documentation/api/latest/guide/configuration.html#using-environment-variables)
        'endpoint_url': str, # AWS endpoint url config param. Mostly for local and development setups
        'parse_body_as_json': bool,
        'flatten_attributes': bool, # Put SQS attributes as first level message
            fields (only works if 'parse_body_as_json' is enabled).
            If not the result dict will contain one field 'message' and
            another 'attributes' with the data inside.
        'delete_messages': bool # Delete messages after processing. Defaults to `true`.
    }

    """

    MAX_BATCH_SIZE = 10 # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sqs.html#SQS.Client.receive_message
    MAX_RETRIES = 3

    BODY_KEY = 'message'
    ATTR_KEY = 'attributes'

    def __init__(self, config: dict = None):
        self.url = None
        self.flatten_attributes = False
        self.parse_body_as_json = False
        self.delete_messages = True

        self._client_id = ''
        self._secret = ''
        self._region = ''
        self._session_token = ''
        self._endpoint_url = ''
        self._sqs_client = None

        if config:
            self.configure(config)

    def configure(self, config: dict):
        self.url = config.get('url')
        if not self.url:
            raise RuntimeError("'url' parameter expected in configuration")

        self.flatten_attributes = config.get('flatten_attributes', self.flatten_attributes)
        self.parse_body_as_json = config.get('parse_body_as_json', self.parse_body_as_json)

        self._client_id = config.get('client_id', self._client_id)
        self._secret = config.get('secret', self._secret)
        self._region = config.get('region', self._region)
        self._session_token = config.get('session_token', self._session_token)
        self._endpoint_url = config.get('endpoint_url', self._endpoint_url)
        self.delete_messages = config.get('delete_messages', self.delete_messages)

    def get_one(self, data: dict = {}) -> dict:
        sqs_req = {
            'QueueUrl': self.url,
            'MessageAttributeNames': ['All'],
            'MaxNumberOfMessages': 1
        }
        sqs_client = self._get_sqs_client()

        res = sqs_client.receive_message(**sqs_req)

        result = self._process_response(res)
        if self.delete_messages:
            self._delete_messages(res)

        if len(result): return result[0]

        return {}

    def get_all(self) -> List[dict]:
        req_attempt_id = str(uuid4())
        res_messages = True
        retries = 0
        sqs_client = self._get_sqs_client()

        result = []
        while res_messages and retries < self.MAX_RETRIES:
            try:
                sqs_res = sqs_client.receive_message(
                    QueueUrl = self.url,
                    MessageAttributeNames = ['All'],
                    MaxNumberOfMessages = self.MAX_BATCH_SIZE,
                    ReceiveRequestAttemptId = req_attempt_id
                )
            except ClientError:
                retries += 1
                continue

            msgs = self._process_response(sqs_res)
            if self.delete_messages:
                self._delete_messages(sqs_res)

            result.extend(msgs)

            res_messages = len(msgs)
            retries = 0
            req_attempt_id = str(uuid4())

        if retries > self.MAX_RETRIES:
            self.logger.warn('Finished processing because max retries reached')

        return result

    def _process_response(self, sqs_res: dict) -> List[dict]:
        proc_res = []
        if not 'Messages' in sqs_res: return []

        for msg in sqs_res['Messages']:
            proc_res.append(self._process_msg(msg))

        return proc_res

    def _delete_messages(self, sqs_res: dict):
        sqs_client = self._get_sqs_client()
        if not 'Messages' in sqs_res: return

        for msg in sqs_res['Messages']:
            sqs_client.delete_message(
                QueueUrl = self.url,
                ReceiptHandle = msg['ReceiptHandle']
            )

    def _process_msg(self, msg: dict) -> dict:
        if not self.parse_body_as_json:
            return {
                self.BODY_KEY: msg['Body'],
                self.ATTR_KEY: self._process_attributes(msg['MessageAttributes'])
            }

        msg_body = json.loads(msg['Body'])
        msg_attr = self._process_attributes(msg['MessageAttributes'])

        proc_msg = {}
        if self.flatten_attributes:
            proc_msg = {**msg_body, **msg_attr}
        else:
            proc_msg = {
                self.BODY_KEY: msg_body,
                self.ATTR_KEY: msg_attr
            }

        return proc_msg

    def _process_attributes(self, attributes: Dict[str, dict]) -> Dict[str, Any]:
        proc_attr = {}
        for attr_name, attr_data in attributes.items():
            attr_val = None
            if attr_data['DataType'] in ['String', 'Number']:
                attr_val = attr_data['StringValue']
            else:
                attr_val = attr_data['BinaryValue']

            proc_attr[attr_name] = attr_val

        return proc_attr

    def _get_sqs_client(self):
        if self._sqs_client:
            return self._sqs_client

        config = {}
        if self._client_id and self._secret:
            config['aws_access_key_id'] = self._client_id
            config['aws_secret_access_key'] = self._secret

        if self._region:
            config['region_name'] = self._region

        if self._session_token:
            config['aws_session_token'] = self._session_token

        if self._endpoint_url:
            config['endpoint_url'] = self._endpoint_url

        self._sqs_client = boto3.client('sqs', **config)

        return self._sqs_client



import re
import json
from typing import List
from requests import request

from . import IDataOutput


class RequestDataOutput(IDataOutput):
    """ DataOutput that performs requests. Config dict format:
    {
        'url': str, # Will interpolate data variables in url when is called by
            migration using "{var_name}" synthax
        'query_params', # Name of variables that must be used as query params
            in request
        'req_verb': str, # Default 'POST'
        'headers': dict,
        'encode_json': bool # Encode data dictionary to JSON. Default `False`
        'json_response': bool # Parse JSON response to object. Default `False`
    }
    """

    URL_PARSE_RX = re.compile('{([^}]*)}') # type: re.Pattern

    def __init__(self, config: dict = None):
        self._config = config

        self._url = ''
        self._query_params = []
        self._req_verb = 'POST'
        self._headers = {}

        self._encode_json = False
        self._json_response = False
        self._url_vars = []

        if config: self.configure(config)

    def configure(self, config: dict):
        self._url = config.get('url')
        if not self._url:
            raise RuntimeError("'url' paramter needed in configuration")

        self._req_verb = config.get('req_verb', self._req_verb)
        self._query_params = config.get('query_params', self._query_params)
        self._headers = config.get('headers', self._headers)

        self._json_response = config.get('json_response', self._json_response)
        self._encode_json = config.get('encode_json', self._encode_json)

        self._process_url()

    def send_one(self, data: dict) -> dict:
        """ Sends requests and return a dict with fields:
         - `code`: Response code
         - `message`: Content of the response (dict if parse JSON is activated,
           bytes otherwise)

        Args:
            data (dict): Payload data.

        Returns:
            dict: Response data (code and message).
        """
        payload = data.copy()
        url = self._generate_url(payload)
        q_params = self._get_query_params(payload)

        if self._encode_json:
            payload = json.dumps(payload)

        res = request(self._req_verb, url, headers=self._headers,
                      params=q_params, data=payload)

        msg = res.content
        if self._json_response:
            msg = res.json()

        return {'code': res.status_code, 'message': msg}

    def send_all(self, data: List[dict]) -> List[dict]:
        """ Sends are request per dict mesasge in list. Returns a list of
        responses as specified in `self.send_one`.

        Args:
            data (List[dict]): List of payloads for requests.

        Returns:
            List[dict]: Responses data (code and message per request).
        """
        return [self.send_one(d) for d in data]

    def _generate_url(self, data: dict) -> str:
        """ Generate url interpolating variables.

        Args:
            data (dict): Data where the variables values are searched

        Returns:
            str: Url
        """
        if not self._url_vars: return self._url

        try:
            var_map = {var:data[var] for var in self._url_vars}

            url = self._url
            for var, val in var_map.items():
                url = re.sub(f'{{{var}}}', val, url)

                # For now we are not going to sendig values that are used in
                # url in payload
                del(data[var])

            return url
        except KeyError as ke:
            raise RuntimeError('Value for url variable not found', data) from ke

    def _get_query_params(self, data: dict) -> dict:
        q_params = {}
        for var in self._query_params:
            val = data.get(var)
            if val is None: continue

            q_params[var] = val

            # For now we are not going to sendig values that are used in
            # url in payload
            del(data[var])

        return q_params

    def _process_url(self):
        self._url_vars = self.URL_PARSE_RX.findall(self._url)
